# slimjim.util
import subprocess

# ===========================================================================

def common_args(parser):
    parser.add_argument('-r', '--raw', action='store_true', help=("Type exactly "
        "what is seen, don't compensate for auto-indent in the target"))
    parser.add_argument('app', help=('Name of app to target keystrokes to'))


def validate_app(target_app):
    proc = subprocess.run(['/usr/local/bin/sendkeys', 'apps'],
        capture_output=True)

    if not target_app in str(proc.stdout):
        print((f'Target app "{target_app}" not found by sendkeys, run '
            '"sendkeys apps" to see a full list of possible targets'))
        quit()



# -*- coding: utf-8 -*-

######
# This file is a copy of a future version of asciimatics's TextBox, one the
# package gets upgraded this can be removed
######

"""This module implements a multi line editing text box"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from builtins import chr
from builtins import str
from copy import copy
from logging import getLogger
from asciimatics.event import KeyboardEvent, MouseEvent
from asciimatics.screen import Screen
from asciimatics.strings import ColouredText
from asciimatics.widgets.widget import Widget
from asciimatics.widgets.utilities import _find_min_start, _enforce_width, _get_offset

# Logging
logger = getLogger(__name__)


class TextBox(Widget):
    """
    A TextBox is a widget for multi-line text editing.

    It consists of a framed box with option label.
    """

    __slots__ = ["_label", "_line", "_column", "_start_line", "_start_column", "_required_height",
                 "_as_string", "_line_wrap", "_on_change", "_reflowed_text_cache", "_parser",
                 "_readonly", "_hide_cursor", "_auto_scroll"]

    def __init__(self, height, label=None, name=None, as_string=False, line_wrap=False, parser=None,
                 on_change=None, readonly=False, **kwargs):
        """
        :param height: The required number of input lines for this TextBox.
        :param label: An optional label for the widget.
        :param name: The name for the TextBox.
        :param as_string: Use string with newline separator instead of a list
            for the value of this widget.
        :param line_wrap: Whether to wrap at the end of the line.
        :param parser: Optional parser to colour text.
        :param on_change: Optional function to call when text changes.
        :param readonly: Whether the widget prevents user input to change values.  Default is False.

        Also see the common keyword arguments in :py:obj:`.Widget`.
        """
        super(TextBox, self).__init__(name, **kwargs)
        self._label = label
        self._line = 0
        self._column = 0
        self._start_line = 0
        self._start_column = 0
        self._required_height = height
        self._as_string = as_string
        self._line_wrap = line_wrap
        self._parser = parser
        self._on_change = on_change
        self._reflowed_text_cache = None
        self._readonly = readonly
        self._hide_cursor = False
        self._auto_scroll = True

    def update(self, frame_no):
        self._draw_label()

        # Calculate new visible limits if needed.
        height = self._h
        if not self._line_wrap:
            self._start_column = min(self._start_column, self._column)
            self._start_column += _find_min_start(
                str(self._value[self._line][self._start_column:self._column + 1]),
                self.width,
                self._frame.canvas.unicode_aware,
                self._column >= self.string_len(str(self._value[self._line])))

        # Clear out the existing box content
        (colour, attr, background) = self._pick_colours("readonly" if self._readonly else "edit_text")
        self._frame.canvas.clear_buffer(
            colour, attr, background, self._x + self._offset, self._y, self.width, height)

        # Convert value offset to display offsets
        # NOTE: _start_column is always in display coordinates.
        display_text = self._reflowed_text
        display_start_column = self._start_column
        display_line, display_column = 0, 0
        for i, (_, line, col) in enumerate(display_text):
            if line < self._line or (line == self._line and col <= self._column):
                display_line = i
                display_column = self._column - col

        # Restrict to visible/valid content.
        self._start_line = max(0, max(display_line - height + 1,
                                      min(self._start_line, display_line)))

        # Render visible portion of the text.
        for line, (text, _, _) in enumerate(display_text):
            if self._start_line <= line < self._start_line + height:
                paint_text = _enforce_width(
                    text[display_start_column:], self.width, self._frame.canvas.unicode_aware)
                self._frame.canvas.paint(
                    str(paint_text),
                    self._x + self._offset,
                    self._y + line - self._start_line,
                    colour, attr, background,
                    colour_map=paint_text.colour_map if hasattr(paint_text, "colour_map") else None)

        # Since we switch off the standard cursor, we need to emulate our own
        # if we have the input focus.
        if self._has_focus and not self._hide_cursor:
            line = str(display_text[display_line][0])
            logger.debug("Cursor: %d,%d", display_start_column, display_column)
            text_width = self.string_len(line[display_start_column:display_column])

            self._draw_cursor(
                " " if display_column >= len(line) else line[display_column],
                frame_no,
                self._x + self._offset + text_width,
                self._y + display_line - self._start_line)

    def reset(self):
        # Reset to original data and move to end of the text.
        self._start_line = 0
        self._start_column = 0
        if self._auto_scroll or self._line > len(self._value) - 1:
            self._line = len(self._value) - 1

        self._column = 0 if self._is_disabled else len(self._value[self._line])
        self._reflowed_text_cache = None

    def _change_line(self, delta):
        """
        Move the cursor up/down the specified number of lines.

        :param delta: The number of lines to move (-ve is up, +ve is down).
        """
        # Ensure new line is within limits
        self._line = min(max(0, self._line + delta), len(self._value) - 1)

        # Fix up column if the new line is shorter than before.
        if self._column >= len(self._value[self._line]):
            self._column = len(self._value[self._line])

    def process_event(self, event):
        def _join(a, b):
            if self._parser:
                return ColouredText(a, self._parser, colour=b[0].first_colour).join(b)
            return a.join(b)

        if isinstance(event, KeyboardEvent):
            old_value = copy(self._value)
            if event.key_code in [10, 13] and not self._readonly:
                # Split and insert line  on CR or LF.
                self._value.insert(self._line + 1,
                                   self._value[self._line][self._column:])
                self._value[self._line] = self._value[self._line][:self._column]
                self._line += 1
                self._column = 0
            elif event.key_code == Screen.KEY_BACK and not self._readonly:
                if self._column > 0:
                    # Delete character in front of cursor.
                    self._value[self._line] = _join(
                        "",
                        [self._value[self._line][:self._column - 1], self._value[self._line][self._column:]])
                    self._column -= 1
                else:
                    if self._line > 0:
                        # Join this line with previous
                        self._line -= 1
                        self._column = len(self._value[self._line])
                        self._value[self._line] += \
                            self._value.pop(self._line + 1)
            elif event.key_code == Screen.KEY_DELETE and not self._readonly:
                if self._column < len(self._value[self._line]):
                    self._value[self._line] = _join(
                        "",
                        [self._value[self._line][:self._column], self._value[self._line][self._column + 1:]])
                else:
                    if self._line < len(self._value) - 1:
                        # Join this line with next
                        self._value[self._line] += \
                            self._value.pop(self._line + 1)
            elif event.key_code == Screen.KEY_PAGE_UP:
                self._change_line(-self._h)
            elif event.key_code == Screen.KEY_PAGE_DOWN:
                self._change_line(self._h)
            elif event.key_code == Screen.KEY_UP:
                self._change_line(-1)
            elif event.key_code == Screen.KEY_DOWN:
                self._change_line(1)
            elif event.key_code == Screen.KEY_LEFT:
                # Move left one char, wrapping to previous line if needed.
                self._column -= 1
                if self._column < 0:
                    if self._line > 0:
                        self._line -= 1
                        self._column = len(self._value[self._line])
                    else:
                        self._column = 0
            elif event.key_code == Screen.KEY_RIGHT:
                # Move right one char, wrapping to next line if needed.
                self._column += 1
                if self._column > len(self._value[self._line]):
                    if self._line < len(self._value) - 1:
                        self._line += 1
                        self._column = 0
                    else:
                        self._column = len(self._value[self._line])
            elif event.key_code == Screen.KEY_HOME:
                # Go to the start of this line
                self._column = 0
            elif event.key_code == Screen.KEY_END:
                # Go to the end of this line
                self._column = len(self._value[self._line])
            elif event.key_code >= 32 and not self._readonly:
                # Insert any visible text at the current cursor position.
                self._value[self._line] = _join(
                    chr(event.key_code),
                    [self._value[self._line][:self._column], self._value[self._line][self._column:]])
                self._column += 1
            else:
                # Ignore any other key press.
                return event

            # If we got here we might have changed the value...
            if old_value != self._value:
                self._reflowed_text_cache = None
                if self._on_change:
                    self._on_change()

        elif isinstance(event, MouseEvent):
            # Mouse event - rebase coordinates to Frame context.
            if event.buttons != 0:
                if self.is_mouse_over(event, include_label=False):
                    # Find the line first.
                    clicked_line = event.y - self._y + self._start_line
                    if self._line_wrap:
                        # Line-wrapped text needs to be mapped to visible lines
                        display_text = self._reflowed_text
                        clicked_line = min(clicked_line, len(display_text) - 1)
                        text_line = display_text[clicked_line][1]
                        text_col = display_text[clicked_line][2]
                    else:
                        # non-wrapped just needs a little end protection
                        text_line = max(0, clicked_line)
                        text_col = 0
                    self._line = min(len(self._value) - 1, text_line)

                    # Now figure out location in text based on width of each glyph.
                    self._column = (self._start_column + text_col +
                                    _get_offset(
                                        str(self._value[self._line][self._start_column + text_col:]),
                                        event.x - self._x - self._offset,
                                        self._frame.canvas.unicode_aware))
                    self._column = min(len(self._value[self._line]), self._column)
                    self._column = max(0, self._column)
                    return None
            # Ignore other mouse events.
            return event
        else:
            # Ignore other events
            return event

        # If we got here, we processed the event - swallow it.
        return None

    def required_height(self, offset, width):
        return self._required_height

    @property
    def _reflowed_text(self):
        """
        The text as should be formatted on the screen.

        This is an array of tuples of the form (text, value line, value column offset) where
        the line and column offsets are indeces into the value (not displayed glyph coordinates).
        """
        if self._reflowed_text_cache is None:
            if self._line_wrap:
                self._reflowed_text_cache = []
                limit = self._w - self._offset
                for i, line in enumerate(self._value):
                    column = 0
                    while self.string_len(str(line)) >= limit:
                        sub_string = _enforce_width(
                            line, limit, self._frame.canvas.unicode_aware)
                        self._reflowed_text_cache.append((sub_string, i, column))
                        line = line[len(sub_string):]
                        column += len(sub_string)
                    self._reflowed_text_cache.append((line, i, column))
            else:
                self._reflowed_text_cache = [(x, i, 0) for i, x in enumerate(self._value)]

        return self._reflowed_text_cache

    @property
    def hide_cursor(self):
        """
        Set to True to stop the cursor from showing.  Defaults to False.
        """
        return self._hide_cursor

    @hide_cursor.setter
    def hide_cursor(self, new_value):
        self._hide_cursor = new_value

    @property
    def auto_scroll(self):
        """
        When set to True the TextBox will scroll to the bottom when created or
        next text is added. When set to False, the current scroll position
        will remain even if the contents are changed.

        Defaults to True.
        """
        return self._auto_scroll

    @auto_scroll.setter
    def auto_scroll(self, new_value):
        self._auto_scroll = new_value

    @property
    def value(self):
        """
        The current value for this TextBox.
        """
        if self._value is None:
            self._value = [""]
        return "\n".join([str(x) for x in self._value]) if self._as_string else self._value

    @value.setter
    def value(self, new_value):
        # Convert to the internal format
        old_value = self._value
        if new_value is None:
            new_value = [""]
        elif self._as_string:
            new_value = new_value.split("\n")
        self._value = new_value

        # TODO: Sort out speed of this code
        if self._parser:
            new_value = []
            last_colour = None
            for line in self._value:
                if hasattr(line, "raw_text"):
                    value = line
                else:
                    value = ColouredText(line, self._parser, colour=last_colour)
                new_value.append(value)
                last_colour = value.last_colour
            self._value = new_value
        self.reset()

        # Only trigger the notification after we've changed the value.
        if old_value != self._value and self._on_change:
            self._on_change()

    @property
    def readonly(self):
        """
        Whether this widget is readonly or not.
        """
        return self._readonly

    @readonly.setter
    def readonly(self, new_value):
        self._readonly = new_value

    @property
    def frame_update_count(self):
        # Force refresh for cursor if needed.
        if self._has_focus and not self._frame.reduce_cpu and not self._hide_cursor:
            return 5

        return 0



#!/usr/bin/env python
import subprocess

from asciimatics.widgets import Button, Frame, Label, Layout
from asciimatics.exceptions import StopApplication

from slimjim.codebox import CodeBox

# ===========================================================================

class FileFrame(Frame):
    def __init__(self, screen, target_app, filename, raw_mode):
        with open(filename) as f:
            content = f.read()

        self.target_app = target_app
        self.raw_mode = raw_mode

        initial_data = {
            'content':content.split('\n'),
        }

        super(FileFrame, self).__init__(screen, screen.height, screen.width, 
            has_border=False, data=initial_data)

        self.palette['block_box'] = (7, 1, 8)
        self.build_layout()
        self.fix()

    def build_layout(self):
        # Edit Area
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)

        if self.raw_mode:
            layout.add_widget(Label('RAW MODE!!!!'))

        self.codebox = CodeBox(25, "Content:", "content", readonly=True)
        self.codebox.auto_scroll = False
        layout.add_widget(self.codebox)

        # Button Bar
        layout = Layout([1, 1])
        self.add_layout(layout)

        layout.add_widget(Button("Send", self.do_send), 0)
        layout.add_widget(Button("Quit", self.do_quit), 1)

    def do_quit(self):
        raise StopApplication('done' + str(self.data))

    def do_send(self):
        box = self.codebox
        code_line = box._reflowed_text[box._line][0]
        prev_line = ''
        keyset = ''

        if box._line > 1:
            prev_line = box._reflowed_text[box._line - 1][0]

        code_indent = len(code_line) - len(code_line.lstrip())
        prev_indent = len(prev_line) - len(prev_line.lstrip())

        if code_indent > prev_indent:
            # More indent
            keyset += '<c:tab>'
        elif code_indent < prev_indent:
            # Less indent, possibly several times
            outdent = (prev_indent - code_indent) // 4
            for _ in range(0, outdent):
                keyset += '<c:delete>'

        # Type result in actual window
        keyset += code_line.lstrip()
        keyset += '<c:enter>'

        subprocess.run([
            '/usr/local/bin/sendkeys',
            '--application-name',
            self.target_app,
            '--characters',
            f'{keyset}'])

        # Line done, move to next one
        box._change_line(1)
        self.screen.force_update()



# -*- coding: utf-8 -*-
"""This module implements a multi line editing text box"""
from asciimatics.widgets.utilities import _find_min_start, _enforce_width

# When asciimatics gets upgraded use the real TextBox
from slimjim.matics_box import TextBox

class CodeBox(TextBox):
    def update(self, frame_no):
        self._draw_label()

        # Calculate new visible limits if needed.
        height = self._h
        if not self._line_wrap:
            self._start_column = min(self._start_column, self._column)
            self._start_column += _find_min_start(
                str(self._value[self._line][self._start_column:self._column + 1]),
                self.width,
                self._frame.canvas.unicode_aware,
                self._column >= self.string_len(str(self._value[self._line])))

        # Clear out the existing box content
        (colour, attr, background) = self._pick_colours("readonly" if self._readonly else "edit_text")
        self._frame.canvas.clear_buffer(
            colour, attr, background, self._x + self._offset, self._y, self.width, height)

        # Convert value offset to display offsets
        # NOTE: _start_column is always in display coordinates.
        display_text = self._reflowed_text
        display_start_column = self._start_column
        display_line, display_column = 0, 0
        for i, (_, line, col) in enumerate(display_text):
            if line < self._line or (line == self._line and col <= self._column):
                display_line = i
                display_column = self._column - col

        # Restrict to visible/valid content.
        self._start_line = max(0, max(display_line - height + 1,
                                      min(self._start_line, display_line)))

        # Render visible portion of the text.
        for line, (text, _, _) in enumerate(display_text):
            if self._start_line <= line < self._start_line + height:
                paint_text = _enforce_width(
                    text[display_start_column:], self.width, self._frame.canvas.unicode_aware)
                self._frame.canvas.paint(
                    str(paint_text),
                    self._x + self._offset,
                    self._y + line - self._start_line,
                    colour, attr, background,
                    colour_map=paint_text.colour_map if hasattr(paint_text, "colour_map") else None)

        #------ Up to here has been a copy of TextWidget, now it changes ----

        # Since we switch off the standard cursor, we need to emulate our own
        # if we have the input focus.
        if self._has_focus or self._readonly:
            line = str(display_text[display_line][0])
            text_width = self.string_len(
                line[display_start_column:display_column])

            if self._readonly:
                padding = (self.width - text_width) * ' '
                self._draw_cursor(
                    line + padding,
                    0,
                    self._x + self._offset,
                    self._y + display_line - self._start_line)
            else:
                self._draw_cursor(
                    " " if display_column >= len(line) else line[display_column],
                    frame_no,
                    self._x + self._offset + text_width,
                    self._y + display_line - self._start_line)

    def reset_cursor(self):
        self._change_line(-1)

        #self._line = 0
        #self._column = 0
        #self.update(0)
        self._frame.screen.force_update()



# slimjim.spec_frame
import subprocess

from asciimatics.widgets import Frame, Layout, Button, Label
from asciimatics.exceptions import StopApplication

from slimjim.matics_box import TextBox
from slimjim.codebox import CodeBox

# ===========================================================================

#import logging
#logging.basicConfig(filename="debug.log", level=logging.DEBUG)
#logger = logging.getLogger()

# ===========================================================================

class NewContent(Exception):
    pass


class Content:
    def __init__(self):
        self.instructions = ''
        self.block = ''
        self.doing_block = False
        self.base_indent = 0

    def add_line(self, line):
        if self.doing_block:
            if line.startswith('---:'):
                # next instruction line
                raise NewContent()

            # was block line, continue building the block
            self.block += line
        else:
            # currently taking instructions
            if line.startswith('---:'):
                # still taking instructions
                self.instructions += line[4:].strip() + '\n'
            else:
                # switch to block mode
                self.doing_block = True
                self.block = line
                self.base_indent = len(line) - len(self.block)

# ===========================================================================

class SpecFrame(Frame):
    def __init__(self, screen, target_app, specfile, raw_mode):
        self.content = [Content()]
        self.current_item = 0
        self.target_app = target_app
        self.raw_mode = raw_mode

        self.parse_content(specfile)

        # Initial values for the widgets
        initial_data = {
            'instructions':self.content[0].instructions.split('\n'),
            'block':self.content[0].block.split('\n'),
        }

        super(SpecFrame, self).__init__(screen, screen.height, screen.width, 
            has_border=False, data=initial_data)

        self.palette['block_box'] = (7, 1, 8)
        self.build_layout()
        self.fix()

    def build_layout(self):
        # Edit Area
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)

        if self.raw_mode:
            layout.add_widget(Label('RAW MODE!!!'))

        self.instruction_box = TextBox(5, "Instructions:", "instructions", 
            readonly=True)
        self.instruction_box.auto_scroll = False
        layout.add_widget(self.instruction_box)

        self.block_box = CodeBox(20, "Block:", "block", readonly=True)
        self.block_box.custom_colour = 'block_box'
        self.block_box.auto_scroll = False
        layout.add_widget(self.block_box)

        # Button Bar
        layout = Layout([1, 1, 1, 1, 1])
        self.add_layout(layout)

        self.send_button = Button("Send", self.do_send)
        layout.add_widget(self.send_button, 0)

        self.line_button = Button("Line", self.do_line)
        layout.add_widget(self.line_button, 1)

        layout.add_widget(Button("Back", self.do_back), 2)

        self.next_button = Button("Next", self.do_next)
        layout.add_widget(self.next_button, 3)

        layout.add_widget(Button("Quit", self.do_quit), 4)

    def parse_content(self, specfile):
        with open(specfile) as f:
            for line in f:
                try:
                    self.content[-1].add_line(line)
                except NewContent:
                    self.content.append(Content())
                    self.content[-1].add_line(line)

    def do_send(self):
        keyset = ''
        skip_indent = True
        indent = self.content[self.current_item].base_indent
        prev_indent = indent
        blocks = self.content[self.current_item].block.split('\n')
        for line in blocks:
            if skip_indent:
                # first line of code (blanks don't count) is meant to have the
                # target cursor positioned, skip indenting otherwise
                if line.lstrip():
                    # found some code, next line uses indent
                    skip_indent = False

                    # set indent value so that prev_indent is set correctly
                    # at the end of this loop
                    indent = len(line) - len(line.lstrip())
                else:
                    # blank line, still haven't found code, just hit enter
                    keyset += '<c:enter>'
            elif not line.lstrip():
                # line is blank, leave the auto-indent alone, just hit enter
                keyset += '<c:enter>'
            else:
                # subsequent line: manage auto-indent based on previous indent
                keyset += '<c:enter>'
                indent = len(line) - len(line.lstrip())

                if not self.raw_mode:
                    if indent > prev_indent:
                        # More indent
                        keyset += '<c:tab>'
                    elif indent < prev_indent:
                        # Less indent, possibly several times
                        outdent = (prev_indent - indent) // 4
                        for _ in range(0, outdent):
                            keyset += '<c:delete>'

            # Add line to key stroke string
            if self.raw_mode:
                keyset += line
            else:
                keyset += line.lstrip()

            prev_indent = indent

        # Type the block
        subprocess.run([
            '/usr/local/bin/sendkeys',
            '--application-name',
            self.target_app,
            '--characters',
            f'{keyset}'])

        # Advance to the next block
        self.do_next()

    def do_next(self):
        self.current_item += 1
        if self.current_item >= len(self.content):
            self.send_button.disabled = True
            self.next_button.disabled = True
            self.instruction_box.value = ['', ]
            self.block_box.value = ['', ]
        else:
            item = self.content[self.current_item]
            self.instruction_box.value = item.instructions.split('\n')
            self.block_box.value = item.block.split('\n')

        self.screen.force_update()

    def do_line(self):
        # Sends only the highlighted line, ignoring any indent
        box = self.block_box
        code_line = box._reflowed_text[box._line][0]
        keyset = code_line.lstrip() + '<c:enter>'

        subprocess.run([
            '/usr/local/bin/sendkeys',
            '--application-name',
            self.target_app,
            '--characters',
            f'{keyset}'])

        # Line done, move to next one
        box._change_line(1)
        self.screen.force_update()

    def do_back(self):
        self.current_item -= 1
        if self.current_item < 0:
            self.current_item = 0

        item = self.content[self.current_item]
        self.instruction_box.value = item.instructions.split('\n')
        self.block_box.value = item.block.split('\n')
        self.send_button.disabled = False
        self.next_button.disabled = False
        self.screen.force_update()

    def do_quit(self):
        raise StopApplication('done' + str(self.data))



import os, sys, re

from slimjim import __version__

readme = os.path.join(os.path.dirname(__file__), 'README.md')
long_description = open(readme).read()


SETUP_ARGS = dict(
    name='slimjim',
    version=__version__,
    description=('Hackish keystroke injection tool '),
    long_description=long_description,
    url='https://github.com/cltrudeau/slimjim',
    author='Christopher Trudeau',
    author_email='ctrudeau+pypi@arsensa.com',
    license='MIT',
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.10',
    ],
    py_modules = ['slimjim',],
    scripts=['bin/slimfile', 'bin/slimspec', ],
    install_requires = [
        'asciimatics>=1.13.0',
    ],
)

if __name__ == '__main__':
    from setuptools import setup, find_packages

    SETUP_ARGS['packages'] = find_packages()
    setup(**SETUP_ARGS)
